@@ -10,7 +10,7 @@ use HTML::Selector::XPath qw/selector_to_xpath/;
 use Carp qw/ confess /;
 use JSON qw/ decode_json /;
 
-our $VERSION = "0.01";
+our $VERSION = "0.02";
 our @EXPORT = qw/ j /;
 
 use constant {
@@ -29,9 +29,14 @@ my $data = {};
 
 sub new {
     my ($class, $stuff, $before) = @_;
-    $class = ref $class if ref $class;
+    my ($self, $existing_document, $nodes);
 
-    my $nodes;
+    # instance method, reuse document
+    if (ref $class) {
+        $self = $class;
+        $class = ref $self;
+        $existing_document = $self->{document};
+    }
 
     if (blessed $stuff) {
 
@@ -57,19 +62,42 @@ sub new {
         } else {
             $nodes = [ _parse_html($stuff) ];
         }
+
+        # # if called as instance method, import nodes to our document
+        # if (ref $self && defined $self->{nodes}[0]) {
+        #
+        #     my $doc = $self->{nodes}[0]->ownerDocument;
+        #     $doc->adoptNode($_)
+        #         for grep { $_->nodeType != XML_DOCUMENT_NODE } @$nodes;
+        # }
     }
 
-    if (@$nodes) {
-
-        # increment document data refcount
-        my $doc_id = $nodes->[0]->ownerDocument->unique_key;
-        $data->{$doc_id}{refcount}++;
-        # warn sprintf "[Document: %s] incrementing Q1::jQuery data ref count: %d\n", $doc_id, $data->{$doc_id}{refcount};
-
-    }
-
+    # catch bugs :)
     confess "undefined node" if grep { !defined } @$nodes;
-    bless({ nodes => $nodes, before => $before }, $class);
+
+    # adopt nodes to existing document
+    if ($existing_document) {
+
+        my $doc_id = $existing_document->unique_key;
+        $existing_document->adoptNode($_)
+            for grep { $_->ownerDocument->unique_key != $doc_id }
+                grep { $_->nodeType != XML_DOCUMENT_NODE }
+                @$nodes;
+    }
+
+    my $document = defined $nodes->[0] ? $nodes->[0]->ownerDocument
+                                       : XML::LibXML->createDocument;
+
+    # increment document data refcount
+    my $doc_id = $document->unique_key;
+    $data->{$doc_id}{refcount}++;
+    # printf STDERR "[%s] incremented document %d data ref count: %d\n", __PACKAGE__, $doc_id, $data->{$doc_id}{refcount};
+
+    bless {
+        document => $document,
+        nodes => $nodes,
+        before => $before
+    }, $class;
 }
 
 
@@ -143,7 +171,7 @@ sub end {
 
 sub document {
     my $self = shift;
-    $self->new([ $self->{nodes}[0] ? $self->{nodes}[0]->ownerDocument : () ], $self);
+    $self->new([ $self->{document} ], $self);
 }
 
 sub tagname {
@@ -183,16 +211,20 @@ sub as_html {
         # TODO benchmark as_html() using can() vs nodeType to detect document nodes
         # best method, but only document nodes can toStringHTML()
         if ($_->can('toStringHTML')) {
+            # printf STDERR "%s: toStringHTML\n", ref $_;
             $output .= $_->toStringHTML;
             next;
         }
 
+
         # second best is to call toStringC14N(1), which generates valid HTML (eg. no auto closed <div/>),
         # but dies on some cases with "Failed to convert doc to string in doc->toStringC14N" error.
         # so we fallback to toString()
+        # the most common case where toStringC14N fails is unbound nodes (getOwner returns a DocumentFragment)
         {
             local $@; # protect existing $@
             my $html = eval { $_->toStringC14N(1) };
+            # printf STDERR "%s: %s\n", ref $_->getOwner, ($@ ? "toString: $@" : 'toStringC14N');
             $output .= $@ ? $_->toString : $html;
         }
     }
@@ -338,11 +370,21 @@ sub _node_matches {
     0;
 }
 
+# TODO add() can ruin our refcount for data()
 sub add {
-    my ($self, $stuff) = @_;
-    my $nodes = $self->new($stuff)->{nodes};
-    push @{$self->{nodes}}, @$nodes;
-    $self;
+    my ($self, $stuff, $context) = @_;
+    $context ||= $self->document;
+
+    # find(): add(selector[, context])
+    # new: add(html), add(elements), add(jQuery)
+    my $new_selection = !ref $stuff && $stuff !~ /<(?!!).*?>/
+                            ? $context->find($stuff)
+                            : $self->new($stuff);
+
+    # prepend our nodes
+    unshift @{$new_selection->{nodes}}, @{ $self->{nodes} };
+
+    $new_selection;
 }
 
 sub each {
@@ -367,7 +409,7 @@ sub append {
 
 sub append_to {
     my $self = shift;
-    _append_to($self->{nodes}, $self->new(@_)->{nodes});
+    _append_to($self->{nodes}, (ref $self)->new(@_)->{nodes});
     $self;
 }
 
@@ -404,7 +446,7 @@ sub prepend {
 
 sub prepend_to {
     my $self = shift;
-    _prepend_to($self->{nodes}, $self->new(@_)->{nodes});
+    _prepend_to($self->{nodes}, (ref $self)->new(@_)->{nodes});
     $self;
 }
 
@@ -448,7 +490,7 @@ sub before {
 
 sub insert_before {
     my $self = shift;
-    _insert_before($self->{nodes}, $self->new(@_)->{nodes});
+    _insert_before($self->{nodes}, (ref $self)->new(@_)->{nodes});
     $self;
 }
 
@@ -494,7 +536,7 @@ sub after {
 
 sub insert_after {
     my $self = shift;
-    _insert_after($self->{nodes}, $self->new(@_)->{nodes});
+    _insert_after($self->{nodes}, (ref $self)->new(@_)->{nodes});
     $self;
 }
 
@@ -739,81 +781,112 @@ sub remove_class {
 }
 
 sub data {
-    my ($self, $key, $val) = @_;
 
-    # data()
-    if (!defined $key) {
+    my $self = shift;
 
-        # class method: return whole $data (mainly for test/debug)
-        return $data unless ref $self;
+    # class method: return whole $data (mainly for test/debug)
+    return $data unless ref $self;
 
-        # instance method: return all data from first node
-        my $node = $self->{nodes}[0];
-        return unless $node;
+    # data(key, val)
+    if (@_ == 2 && defined $_[1]) {
 
-        return ($data->{$node->ownerDocument->unique_key}->{$node->unique_key} //= {});
-    }
-
-    # no nodes
-    return $self unless $self->{nodes}->[0];
-
-    # data( obj )
-    if (ref $key eq 'HASH') {
-        # TODO die if not a HASH ref
-
-        foreach my $node (@{$self->{nodes}}) {
-            my $node_data = ($data->{$node->ownerDocument->unique_key}->{$node->unique_key} //= {});
-            $node_data->{$_} = $key->{$_} for keys %$key;
-        }
+        $data->{$_->ownerDocument->unique_key}->{$_->unique_key}->{$_[0]} = $_[1]
+            foreach @{$self->{nodes}};
 
         return $self;
     }
 
-    # data( key )
-    if (! defined $val) {
-        my $node = $self->{nodes}->[0];
-        my $node_data = ($data->{$node->ownerDocument->unique_key}->{$node->unique_key} //= {});
 
-        return $node_data->{$key} if defined $node_data->{$key};
+    if (@_ == 1) {
+
+        # no nodes
+        return unless defined $self->{nodes}->[0];
+
+        # data(undefined)
+        return $self unless defined $_[0];
+
+        # data(obj)
+        if (ref $_[0]) {
+
+            die 'data(obj) only accepts a hashref' unless ref $_[0] eq 'HASH';
+
+            $data->{$_->ownerDocument->unique_key}->{$_->unique_key} = $_[0]
+                foreach @{$self->{nodes}};
+
+            return $self;
+        }
+
+        # data(key)
+        my $key = $_[0];
+        my $node = $self->{nodes}->[0];
+
+        $data->{$node->ownerDocument->unique_key}->{$node->unique_key} = {}
+            unless $data->{$node->ownerDocument->unique_key}->{$node->unique_key};
+
+        my $node_data = $data->{$node->ownerDocument->unique_key}->{$node->unique_key};
 
         # try to pull from data-* attribute
         my $data_attr = 'data-'._decamelize($key);
         $data_attr =~ tr/_/-/;
 
-        if ($node->nodeType == XML_ELEMENT_NODE && defined ($val = $node->getAttribute($data_attr))) {
-
-            # convert
-
-            # number
-            if ($val =~ /^\d+$/) {
-                $val += 0
-            }
-            # json array or object
-            elsif ($val =~ /^(?:\{|\[)/) {
-                {
-                    local $@;
-                    my $data = eval { decode_json $val };
-                    $val = $data unless $@;
-                }
-            }
-            # TODO boolean
-
-            # save on data hash for next time
-            $node_data->{$key} = $val;
-        }
+        $node_data->{$key} = _convert_data_attr_value($node->getAttribute($data_attr))
+            if !exists $node_data->{$key}
+                && $node->nodeType == XML_ELEMENT_NODE
+                && $node->hasAttribute($data_attr);
 
         return $node_data->{$key};
     }
 
-    # data( key, val)
-    foreach my $node (@{$self->{nodes}}) {
-        my $node_data = ($data->{$node->ownerDocument->unique_key}->{$node->unique_key} //= {});
-        $node_data->{$key} = $val;
+    # data()
+    if (@_ == 0) {
+
+        # return all data for first node
+        my $node = $self->{nodes}[0];
+        return unless $node;
+
+        # poor man's //= {} (for perls < 5.10)
+        exists $data->{$node->ownerDocument->unique_key}->{$node->unique_key}->{autovivify_hash};
+        my $node_data = $data->{$node->ownerDocument->unique_key}->{$node->unique_key};
+
+        # pull data-* attributes
+        foreach my $attr (grep { $_->name =~ /^data-/ } $node->attributes) {
+
+            my $key = substr($attr->name, 5);
+            $key =~ tr/-/_/;
+            $key = lcfirst _camelize($key);
+
+            next if exists $node_data->{$key};
+            $node_data->{$key} = _convert_data_attr_value($attr->value);
+        }
+
+        return $node_data;
     }
 
-    return $self;
+    $self;
 }
 
+sub _convert_data_attr_value {
+
+    # number
+    return $_[0] += 0
+        if $_[0] =~ /^\d+$/;
+
+    # json array or object
+    return decode_json($_[0])
+        if  $_[0] =~ /^(?:\{|\[)/;
+
+    # boolean
+    return JSON::true  if $_[0] eq 'true';
+    return JSON::false if $_[0] eq 'false';
+
+    # undef
+    return undef if $_[0] eq 'null' || $_[0] eq 'undefined';
+
+    # other stuff, return unmodified
+    $_[0];
+}
+
+
 
 sub _decamelize
 {
@@ -828,6 +901,11 @@ sub _decamelize
         $s;
 }
 
+sub _camelize {
+        my $s = shift;
+        join('', map{ ucfirst $_ } split(/(?<=[A-Za-z])_(?=[A-Za-z])|\b/, $s));
+}
+
 # decrement data ref counter, delete data when counter == 0
 sub DESTROY {
     my $self = shift;
@@ -838,7 +916,7 @@ sub DESTROY {
     # decrement $data refcount
     my $doc_id = $node->ownerDocument->unique_key;
     $data->{$doc_id}{refcount}--;
-    # warn sprintf "[Document: %s] decrementing Q1::jQuery data ref count: %d\n", $doc_id, $data->{$doc_id}{refcount};
+    # printf STDERR "[%s] decremented document %d data ref count: %d\n", __PACKAGE__, $doc_id, $data->{$doc_id}{refcount};
 
     # delete document data if refcount is 0
     delete $data->{$doc_id}
@@ -931,6 +1009,24 @@ Unless otherwise noted, all methods behave exactly like the javascript version.
 
 =head2 add
 
+Implemented signatures:
+
+=over
+
+=item B<add>(Str B<selector>)
+
+=item B<add>(Str B<selector>, XML::LibXML::jQuery B<context>)
+
+=item B<add>(Str B<html>)
+
+=item B<add>(ArrayRef[L<XML::LibXML::Node>] B<elements>)
+
+=item B<add>(XML::LibXML::jQuery B<selection>)
+
+=back
+
+Documentation and examples at L<http://api.jquery.com/add/>.
+
 =head2 add_class
 
 =head2 after
@@ -953,6 +1049,20 @@ Unless otherwise noted, all methods behave exactly like the javascript version.
 
 =head2 data
 
+Implemented signatures:
+
+=over
+
+=item B<data>(Str B<key>, Any B<value>)
+
+=item B<data>(Str B<key>)
+
+=item B<data>(HashRef B<obj>)
+
+=back
+
+Documentation and examples at L<http://api.jquery.com/data/>.
+
 =head2 detach
 
 =head2 document
